;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "acp" "0.12.2"
  "An ACP (Agent Client Protocol) implementation."
  '((emacs "28.1"))
  :url "https://github.com/xenodium/acp.el"
  :commit "c8ee1d7f70105fba8efa964ca63f38ca94a1e759"
  :revdesc "c8ee1d7f7010")
